''==============================================================================
'' COMMON MODULE - REGISTRY AND REQUEST PRIMITIVES
''==============================================================================
CON
'
' Maximum number of services:
'
MAX_SERVICE     = 96                  ' must be a multiple of 2
'
' Runtime Hub RAM Layout:
'
HUB_SIZE        = $8000               ' size of Hub RAM
FREE_MEM        = HUB_SIZE - 4        ' pointer to lowest used Hub RAM
REGISTRY        = FREE_MEM - 8*4*1    ' start of Registry (one long per cog up, one word per service down)        
REQUESTS        = REGISTRY - 2*MAX_SERVICE - 8*4*2  ' start of Comms Blocks (2 longs per cog)        
RESERVED        = REQUESTS            ' allocate from here down for buffers at runtime

CON
''==============================================================================
'' LAYER 1 METHODS
''==============================================================================
PUB InitializeRegistry | i
'
' this method should be called by the first Spin program executed,
' before it starts or initializes any of the plugins to be loaded.
'
' Note we use the value $FF in the upper byte to mean "unknown" or "none"
'

  ' initialize the FREE_MEM pointer
  long[FREE_MEM] := RESERVED
  ' initialize LAYER 1 registry (cog-oriented registry)
  repeat i from 0 to 7
    long[REGISTRY][i] := REQUESTS + (8*i) + $FF000000
    long[REQUESTS][2 * i] := 0
    long[REQUESTS][2 * i + 1] := 0

  ' initialize LAYER 3 registry (service-oriented registry)
  repeat i from 1 to MAX_SERVICE
     word[REGISTRY][-i] := $FF00


PUB Register (cog, plugin_type)
'
' register a plugin by storing the plugin tyoe in the
' top 8 bits of the cog's registry entry (the lower
' 24 bits are left intact - normally, they point to
' the cog's comms block)
'
  long[REGISTRY][cog] := (plugin_type<<24) + (long[REGISTRY][cog] & $00FFFFFF)

PUB UnRegister (cog)
'
' this method can be used by a plugin to unregister itself
' with the Catalina Kernel. Plugins must unregister themselves
' if they may later be stopped or restarted in a different cog.
' Note we use the value $FF in the upper byte to indicate
' "unknown" or "none"
'
  long[REGISTRY][cog] := (long[REGISTRY][cog] | $FF000000)

PUB Multiple_Register (cogbits, plugin_type) | i
'
' this routine can be used by a plugin to register multiple
' cogs, based on up to 8 bits set in the cogbits parameter
'
   repeat i from 0 to 7
     if cogbits & 1
        Register(i, plugin_type)
     cogbits >>= 1

PUB SendInitializationData(cog, data_1, data_2) : commsblk
'
' this routine can be called to send initialization data
' to a cog. 
'
  commsblk := long[REGISTRY][cog] & $00FFFFFF
  long[commsblk][1] := data_2
  long[commsblk][0] := data_1

PUB SendInitializationDataAndWait(cog, data_1, data_2) : response | commsblk
'
' This routine is similar to SendInitializationData,
' except this routine waits for the initialization 
' to complete and then returns the result.
'
  commsblk := SendInitializationData(cog, data_1, data_2)
  repeat while long[commsblk][0] <> 0
  ' get the response of the request
  return long[commsblk][1]


CON
''==============================================================================
'' LAYER 2 METHODS
''==============================================================================
PUB LocatePlugin(plugin_type) : cog | i
'
' search the registry for the plugin type, returning the cog if found,
' or -1 if no such plugin type is found
'
  repeat i from 0 to 7
    if long[Registry][i] >> 24 == plugin_type
      return i
  return -1
  
PUB WaitForCompletion(cog) : response | commsblk
'
' when the first long of the cog's comms block is zero,
' return the response from the second long
'
  commsblk := long[REGISTRY][cog] & $00FFFFFF
  ' wait for a request to be processed
  repeat while long[commsblk][0] <> 0
  ' get the response of the request
  return long[commsblk][1]

PUB SetRequest(cog, request) | commsblk
'
' Set the first long in the cog's comms block to the specified request.
'
  commsblk := long[REGISTRY][cog] & $00FFFFFF
  long[commsblk][0] := request

PUB GetRequest(cog) : request | commsblk
'
' Get the reqyest from the first long in the v.
'
  commsblk := long[REGISTRY][cog] & $00FFFFFF
  request := long[commsblk][0]

PUB SetResponse(cog, response) | commsblk
'
' Set the second long in the cog's comms block to the specified response.
'
  commsblk := long[REGISTRY][cog] & $00FFFFFF
  long[commsblk][1] := response

PUB GetResponse(cog) : response | commsblk
'
' Get the second long in the cog's comms block.
'
  commsblk := long[REGISTRY][cog] & $00FFFFFF
  ' get the response of a previous request
  return long[commsblk][1]

PUB PerformRequest (cog, request) : response
'
' Set the request in the first long of the comms block of the cog,
' wait till it is acknowledged, then return the response (note that
' this method does not check that the previous request has completed)
' 
  SetRequest(cog, request)
  return WaitForCompletion(cog)

PUB ShortPluginRequest(plugin_type, request_id, parameter) : response | cog, request
'
' this routine attempts to locate a plugin of the specified type, then
' sends a short command made up of the service and the parameter to it,
' then waits for the service to complete and returns the response. It
' returns -1 if it cannnot locate the plugin.
'
  cog := LocatePlugin(plugin_type)
  if (cog => 0)
    return PerformRequest(cog, (request_id << 24) | (parameter & $00FFFFFF))
  return -1

PUB LongPluginRequest(plugin_type, request_id, parameter) : response | cog, request
'
' this routine attempts to locate a plugin of the specified type, then
' sends a long command made up of the service and a pointer to the parameter to it,
' then waits for the service to complete and returns the response. It
' returns -1 if it cannnot locate the plugin.
'
  cog := LocatePlugin(plugin_type)
  if (cog => 0)
    return PerformRequest(cog, (request_id << 24) | @parameter)
  return -1

PUB LongPluginRequest_2(plugin_type, request_id, param_1, param_2) : response | cog, request
'
' this routine attempts to locate a plugin of the specified type, then
' sends a long command made up of the service and a pointer to the parameter to it,
' then waits for the service to complete and returns the response. It
' returns -1 if it cannnot locate the plugin.
'
  cog := LocatePlugin(plugin_type)
  if (cog => 0)
    return PerformRequest(cog, (request_id << 24) | @param_1)
  return -1

CON  
''==============================================================================
' LAYER 3 METHODS
''==============================================================================
PUB RegisterServiceCog(service_id, lock, cog, request_id)
'
' Use this function to register services if you know the cog
'
  if (service_id > 0) and (service_id =< MAX_SERVICE) 
    if (cog => 0) and (cog < 8)
      word[REGISTRY][-service_id] := ((cog&$F)<<12)|((lock&$F)<<8)|(request_id&$FF)
      return cog
  return -1

PUB RegisterServiceType(service_id, lock, plugin_type, request_id) : cog
'
' Use this function to register services if you know the type but not the cog
'
  if (service_id > 0) and (service_id =< MAX_SERVICE) 
    cog := LocatePlugin(plugin_type)
    if (cog => 0) and (cog < 8)
      word[REGISTRY][-service_id] := ((cog&$F)<<12)|((lock&$F)<<8)|(request_id&$FF)
      return cog
  return -1

PUB UnregisterService(service)
'
' Remove details of the service from the registry
'
  if (service > 0) and (service =< MAX_SERVICE)
    word[REGISTRY][-service] := $FF00

PUB ShortServiceRequest(service_id, parameter) : response | entry, lock, cog
'
' Request the specified service using a short request, passing the
' appropriate request code and the lower 24 bits of the parameter
' to the plugin currently servicing these requests. Return -1 on error.
' If the service requires the use of a lock, the lock is set prior to
' the service call, and cleared afterwards.
'
  if (service_id > 0) and (service_id =< MAX_SERVICE)
    entry := word[REGISTRY][-service_id] 
    cog := entry>>12 
    if (cog < 8)
      lock := (entry>>8) & $F
      if (lock < 8)
        repeat until not LOCKSET(lock)
        response := PerformRequest(cog, (entry<<24) | (parameter & $00FFFFFF))
        LOCKCLR(lock)
        return 
      else
        return PerformRequest(cog, (entry<<24) | (parameter & $00FFFFFF))
  return -1
  
PUB LongServiceRequest(service_id, parameter) : response | entry, lock, cog
'
' Request the specified service using a long request, passing the
' appropriate request code and a pointer to the parameter to the
' plugin currently servicing these requests. Return -1 on error.
' If the service requires the use of a lock, the lock is set prior to
' the service call, and cleared afterwards.
'
  if (service_id > 0) and (service_id =< MAX_SERVICE)
    entry := word[REGISTRY][-service_id] 
    cog := entry>>12 
    if (cog < 8)
      lock := (entry>>8) & $F
      if (lock < 8)
        repeat until not LOCKSET(lock)
        response := PerformRequest(cog, (entry<<24) | @parameter)
        LOCKCLR(lock)
        return 
      else
        return PerformRequest(cog, (entry<<24) | @parameter)
  return -1

PUB LongServiceRequest_2(service_id, param_1, param_2) : response | entry, lock, cog
'
' Request the specified service using a long request, passing the
' appropriate request code and a pointer to param_1 and param_2 to
' the plugin currently servicing these requests. Return -1 on error.
' If the service requires the use of a lock, the lock is set prior to
' the service call, and cleared afterwards.
'
  if (service_id > 0) and (service_id =< MAX_SERVICE)
    entry := word[REGISTRY][-service_id] 
    cog := entry>>12 
    if (cog < 8)
      lock := (entry>>8) & $F
      if (lock < 8)
        repeat until not LOCKSET(lock)
        response := PerformRequest(cog, (entry<<24) | @param_1)
        LOCKCLR(lock)
        return 
      else
        return PerformRequest(cog, (entry<<24) | @param_1)
  return -1

PUB ServiceLock(service_id) : lock
'
' Return the lock allocated to the service (or a value => 8 if
' no lock has been allocated). Return -1 on error.
'
  if (service_id > 0) and (service_id =< MAX_SERVICE)
    return (word[REGISTRY][-service_id]>>8)&$F
  return -1

PUB NoLockShortServiceRequest(service_id, parameter) : response | entry, cog
'
' Request the specified service using a short request, passing the
' appropriate request code and the lower 24 bits of the parameter
' to the plugin currently servicing these requests. Return -1 on error.
' NOTE: No locking is performed, even if the service requires it - this
' must be done manually prior to the service request.
'
  if (service_id > 0) and (service_id =< MAX_SERVICE)
    entry := word[REGISTRY][-service_id] 
    cog := entry>>12 
    if (cog < 8)
      return PerformRequest(cog, (entry<<24) | (parameter & $00FFFFFF))
  return -1

PUB NoLockLongServiceRequest(service_id, parameter) : response | entry, cog
'
' Request the specified service using a long request, passing the
' appropriate request code and a pointer to the parameter to the
' plugin currently servicing these requests. Return -1 on error.
' NOTE: No locking is performed, even if the service requires it - this
' must be done manually prior to the service request.
'
  if (service_id > 0) and (service_id =< MAX_SERVICE)
    entry := word[REGISTRY][-service_id] 
    cog := entry>>12 
    if (cog < 8)
      return PerformRequest(cog, (entry<<24) | @parameter)
  return -1

PUB NoLockLongServiceRequest_2(service_id, param_1, param_2) : response | entry, cog
'
' Request the specified service using a long request, passing the
' appropriate request code and a pointer to param_1 and param_2 to 
' the plugin currently servicing these requests. Return -1 on error.
' NOTE: No locking is performed, even if the service requires it - this
' must be done manually prior to the service request.
'
  if (service_id > 0) and (service_id =< MAX_SERVICE)
    entry := word[REGISTRY][-service_id] 
    cog := entry>>12 
    if (cog < 8)
      return PerformRequest(cog, (entry<<24) | @param_1)
  return -1
